/*
 * Copyright 2024 <lhearachel@proton.me>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fnmatch.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/stat.h>

#include <narc/api/pack.h>
#include <narc/defs/narc.h>
#include <narc/defs/vfs.h>

#include "command.h"
#include "strbuild.h"
#include "strutil.h"
#include "strvec.h"

typedef intptr_t ssize_t;

// clang-format off
static const char *tag_line = "narc create - create a NARC from a folder of physical files\n";

static const char *short_usage = "Usage: narc create [-h | --help] [options] DIRECTORY\n";

static const char *notes = ""
    "Files within DIRECTORY will be added to the NARC's virutal filesystem (VFS)\n"
    "in lexicographical order.\n"
    "";

static const char *options = ""
    "Options:\n"
    "  -o, --output <FILE>  Persist the output to FILE. By default, output is\n"
    "                       written to “DIRECTORY.narc”.\n"
    "  --naix               Create a Nitro Archive Index (NAIX) header. The output\n"
    "                       NAIX will have the same file-stem as the output NARC.\n"
    "  --order <FILE>       Read lines from FILE as filepaths in DIRECTORY which\n"
    "                       should be added to the NARC's VFS in non-lexicographical\n"
    "                       order. These files will be added to the NARC before any\n"
    "                       other members within DIRECTORY.\n"
    "  --ignore <FILE>      Read lines from FILE as a list of filepaths and glob-\n"
    "                       patterns which may be present in DIRECTORY but should\n"
    "                       not be added to the output."
    "";

static const char *naix_header = ""
    "/*\n"
    " * THIS FILE WAS AUTOMATICALLY GENERATED BY narc\n"
    " *              DO NOT MODIFY!!!\n"
    " */\n"
    "#ifndef %s\n"
    "#define %s\n"
    "\n"
    "";

static const char *naix_member = "#define %s %d\n";

static const char *naix_footer = "\n#endif /* %s */\n";
// clang-format on

struct options {
    bool naix;
    const char *input;
    const char *output;
    const char *order;
    const char *ignore;
};

static int parse_opts(int *argc, const char ***argv, struct options *opts);
static int pack(struct options *opts);
static struct strvec *build_pack_list(DIR *dir, const char *dir_name, const char *order_fname, const char *ignore_fname);
static struct strbuild *start_index(const char *target_fname, char **out_guard);
static void add_to_index(struct strbuild *index, const char *fname, const size_t i);
static void finish_index(struct strbuild *index, const char *guard);

int create(int argc, const char **argv)
{
    if (argc == 0 || match_either(*argv, "-h", "--help")) {
        printf("%s\n%s\n%s\n%s\n", tag_line, short_usage, notes, options);
        return EXIT_SUCCESS;
    }

    struct options opts = {
        .input = NULL,
        .output = NULL,
        .naix = false,
        .order = NULL,
        .ignore = NULL,
    };
    parse_opts(&argc, &argv, &opts);

    if (argc < 1) {
        fprintf(stderr, "narc create: missing required positional argument DIRECTORY\n");
        fprintf(stderr, "%s\n%s\n", short_usage, options);
        return EXIT_FAILURE;
    }

    opts.input = *argv;
    opts.output = opts.output ? opts.output : basename_extend(opts.input, "narc");
    return pack(&opts);
}

static int parse_opts(int *argc, const char ***argv, struct options *opts)
{
    const char **orig_argv = *argv;

    while (*argc > 1) {
        const char *opt = **argv;
        if (opt[0] != '-' || (opt[1] == '-' && opt[2] == '\0')) {
            break;
        }

        (*argc)--;
        (*argv)++;

        if (match_either(opt, NULL, "--naix")) {
            opts->naix = true;
            continue;
        }

        const char *arg = **argv;
        (*argc)--;
        (*argv)++;

        if (match_either(opt, "-o", "--output")) {
            opts->output = arg;
        } else if (match_either(opt, NULL, "--order")) {
            opts->order = arg;
        } else if (match_either(opt, NULL, "--ignore")) {
            opts->ignore = arg;
        }
    }

    return orig_argv - *argv;
}

static int pack(struct options *opts)
{
    DIR *dir = NULL;
    FILE *fout = NULL, *fnaix = NULL;
    char *cwd = NULL;
    struct narc *narc = NULL;
    struct vfs_pack_ctx *ctx = NULL;
    struct strbuild *index = NULL;
    struct strvec *to_pack = NULL;
    char *naix = strcpy_fext(opts->output, "naix.h");

    dir = opendir(opts->input);
    if (dir == NULL) {
        FAIL("narc create: could not open DIRECTORY “%s”: %s\n", opts->input, strerror(errno));
    }

    cwd = getcwd(cwd, 0);
    if (cwd == NULL) {
        FAIL("narc create: could not get current working directory: %s\n", strerror(errno));
    }

    char guard[256] = {0};
    char *guard_p = &guard[0]; // just to make the compiler happy
    index = start_index(naix, &guard_p);
    to_pack = build_pack_list(dir, opts->input, opts->order, opts->ignore);
    if (to_pack == NULL) {
        fprintf(stderr, "narc create: failure while building packing list");
        goto fail;
    }

    chdir(opts->input);
    ctx = narc_pack_start();
    for (size_t i = 0; i < to_pack->count; i++) {
        FILE *f = fopen(to_pack->s[i], "rb");
        if (f == NULL) {
            FAIL("narc create: error while opening file “%s” for reading: %s\n", to_pack->s[i], strerror(errno));
        }

        fseek(f, 0, SEEK_END);
        size_t fsize = ftell(f);
        fseek(f, 0, SEEK_SET);

        unsigned char *image = malloc(fsize);
        if (image == NULL) {
            FAIL("narc create: error while reading file “%s”: %s\n", to_pack->s[i], strerror(errno));
        }

        fread(image, 1, fsize, f);
        narc_pack_file(ctx, image, fsize);
        add_to_index(index, to_pack->s[i], i);
        fclose(f);
    }
    narc = narc_pack(ctx);
    finish_index(index, guard);
    chdir(cwd);

    if (opts->naix) {
        fnaix = fopen(naix, "w");
        if (fnaix == NULL) {
            FAIL("narc create: error while opening file “%s” for writing: %s\n", naix, strerror(errno));
        }
        fwrite(index->s, 1, index->len, fnaix);
        fclose(fnaix);
    }

    fout = fopen(opts->output, "wb");
    if (fout == NULL) {
        FAIL("narc create: error while opening file “%s” for writing: %s\n", opts->output, strerror(errno));
    }
    fwrite(narc, narc->size, 1, fout);

    free(naix);
    free(cwd);
    free(narc);
    strbuild_del(index);
    strvec_del(to_pack);
    fclose(fout);
    closedir(dir);
    return EXIT_SUCCESS;

fail:
    if (fout) {
        fclose(fout);
    }

    if (fnaix) {
        fclose(fnaix);
    }

    if (dir) {
        closedir(dir);
    }

    strbuild_del(index);
    strvec_del(to_pack);
    narc_pack_halt(ctx);
    free(naix);
    free(cwd);
    free(narc);
    return EXIT_FAILURE;
}

static struct strvec *build_pack_list(DIR *dir, const char *dir_name, const char *order_fname, const char *ignore_fname)
{
    struct strvec *all_files = NULL, *ignored = NULL, *lex_sorted_files = NULL;

    // Build the list of ordered files first. These are never ignored.
    all_files = strvec_new(5000);
    if (strvec_from_file(all_files, order_fname)) {
        goto cleanup_error;
    }

    // Load the list of `ignored` names/patterns. This list is kept separate.
    ignored = strvec_new(100);
    if (strvec_from_file(ignored, ignore_fname)) {
        goto cleanup_error;
    }

    // Iterate over the directory, filtering out ignored entries.
    // `readdir`'s ordering is effectively non-deterministic. Collect all the
    // valid files, sort them, then attach them to the list of all packed files.
    errno = 0;
    struct dirent *entry;
    lex_sorted_files = strvec_new(5000);
    while ((entry = readdir(dir)) != NULL) {
        if (match_either(entry->d_name, ".", "..")) {
            continue;
        }

        if (isdir(dir_name, entry->d_name)) {
            continue;
        }

        // Check the ignored list first, since it should be much smaller
        bool exclude = false;
        for (size_t i = 0; i < ignored->count; i++) {
            if (fnmatch(ignored->s[i], entry->d_name, FNM_PERIOD) == 0) {
                exclude = true;
                break;
            }
        }

        if (exclude) {
            continue;
        }

        // Now check the ordered list, to avoid double includes
        for (size_t i = 0; i < all_files->count; i++) {
            if (strcmp(all_files->s[i], entry->d_name) == 0) {
                exclude = true;
                break;
            }
        }

        if (exclude) {
            continue;
        }

        char *fname = malloc(strlen(entry->d_name) + 1);
        strcpy(fname, entry->d_name);
        fname[strlen(entry->d_name)] = '\0';
        strvec_push(lex_sorted_files, fname);
    }

    qsort(lex_sorted_files->s, lex_sorted_files->count, sizeof(char *), strcmp_q);
    strvec_extend(all_files, lex_sorted_files);

    free(lex_sorted_files->s);
    free(lex_sorted_files);
    strvec_del(ignored);

    return all_files;

cleanup_error:
    strvec_del(all_files);
    strvec_del(ignored);
    return NULL;
}

static struct strbuild *start_index(const char *target_fname, char **out_guard)
{
    char guard[256] = "NARC_";
    size_t i;
    char c, *p = &guard[5];
    char *target_fname_base = basename(target_fname);
    for (i = 0; i < 250 && (c = target_fname_base[i]) != '\0'; i++) {
        if (c == '-' || c == '.' || c == '_') {
            *p = '_';
            p++;
        } else if (islower(c)) {
            *p = toupper(c);
            p++;
        } else if (isupper(c)) {
            *p = c;
            p++;
        }
    }
    *p = '\0';
    strcpy(*out_guard, guard);

    struct strbuild *str = strbuild_new(65536);
    strbuild_sprintf(str, naix_header, guard, guard);
    return str;
}

static void add_to_index(struct strbuild *index, const char *fname, const size_t i)
{
    char def_name[256];
    size_t j;
    char c, *p = def_name;
    for (j = 0; j < 256 && (c = fname[j]) != '\0'; j++, p++) {
        if (ispunct(c)) {
            *p = '_';
        } else {
            *p = c;
        }
    }
    *p = '\0';

    strbuild_sprintf(index, naix_member, def_name, i);
}

static void finish_index(struct strbuild *index, const char *guard)
{
    strbuild_sprintf(index, naix_footer, guard);
}
